// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package backend

import (
	"crypto"
	"crypto/internal/boring/sig"
	"internal/goexperiment"
	"runtime"
	"syscall"
)

func init() {
	if v, _ := envGoFIPS(); v == "1" {
		if runtime.GOOS != "linux" && runtime.GOOS != "windows" {
			panic("FIPS mode requested (GOFIPS=1) but not supported on " + runtime.GOOS)
		}
		if runtime.GOOS == "linux" && !iscgo {
			panic("FIPS mode requested (GOFIPS=1) but not supported with cgo disabled")
		}
	}
}

func envGoFIPS() (string, bool) {
	if v, ok := syscall.Getenv("GOFIPS"); ok {
		return v, true
	}
	// TODO: Decide which environment variable to use.
	// See https://github.com/microsoft/go/issues/397.
	if v, ok := syscall.Getenv("GOLANG_FIPS"); ok {
		return v, true
	}
	return "", false
}

// Unreachable marks code that should be unreachable
// when backend is in use.
func Unreachable() {
	if Enabled {
		panic("cryptobackend: invalid code execution")
	} else {
		// Code that's unreachable is exactly the code
		// we want to detect for reporting standard Go crypto.
		sig.StandardCrypto()
	}
}

// Provided by runtime.crypto_backend_runtime_arg0 to avoid os import.
func runtime_arg0() string

func hasSuffix(s, t string) bool {
	return len(s) > len(t) && s[len(s)-len(t):] == t
}

// UnreachableExceptTests marks code that should be unreachable
// when backend is in use. It panics.
func UnreachableExceptTests() {
	// runtime_arg0 is not supported on windows.
	// We are going through the same code patch on linux,
	// so if we are unintentionally calling an 'unreachable' function,
	// we will catch it there.
	if Enabled && !goexperiment.CNGCrypto {
		name := runtime_arg0()
		// If ran on Windows we'd need to allow _test.exe and .test.exe as well.
		if !hasSuffix(name, "_test") && !hasSuffix(name, ".test") {
			println("cryptobackend: unexpected code execution in", name)
			panic("cryptobackend: invalid code execution")
		}
	}
}

func IsRSAKeySupported(primes int) bool {
	if goexperiment.CNGCrypto {
		return primes == 2
	}
	return true
}

func IsHashSupported(h crypto.Hash) bool {
	if goexperiment.CNGCrypto {
		return h != crypto.MD5SHA1
	}
	return true
}

func IsSaltSupported(salt int) bool {
	if goexperiment.CNGCrypto {
		return salt != 0 // rsa.PSSSaltLengthAuto
	}
	return true
}

func IsP224Supported() bool {
	return !goexperiment.CNGCrypto
}
